--
-- MP3 Player, VHDL version (http://www.wiresncode.com/projects/pjrcmp3/fpga)
-- Copyright (c) 2000-1, PJRC.COM, LLC (paul_at_pjrc.com)
-- Copyright (c) 2002, Tom Parker (tparker_at_netspace.net.au)
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_arith.all;
use IEEE.numeric_std.all;

library unisim;
use unisim.all;

entity main is 
  port(clk : in std_logic;
       -- cpu bus
       cpudata : inout std_logic_vector(7 downto 0);
       cpuaddressmsb : in std_logic_vector(15 downto 8);
       psen : in std_logic;
       ale : in std_logic;
       cpurd: in std_logic;
       cpuwr : in std_logic;
       interrupt : out std_logic;

       -- ide/memory signals
       idedramdata : inout std_logic_vector(15 downto 0);
       idedramaddrout : out std_logic_vector(10 downto 0);
       iderd : out std_logic;
       idewr : out std_logic;
       ideresetpin : out std_logic;
       dramrowaddress : out std_logic_vector(3 downto 0);
       dramcasout : out std_logic;

       -- sta013 signals
       sta013requestin : in std_logic;
       sta013data : out std_logic;
       sta013clock : out std_logic;

       -- misc control
       ideen : in std_logic);
end main;

architecture behavior of main is

  signal cpudatain, cpudataout : std_logic_vector(7 downto 0);
  signal adr_en : std_logic;
  signal cpuaddress : std_logic_vector(15 downto 0);
  signal reset : std_logic;
  signal memoryaddress : std_logic_vector(15 downto 0);
  signal drampagedata : std_logic_vector(15 downto 0);
  signal dramwetdo : std_logic;
  signal dmaparamdatain : std_logic_vector(15 downto 0);
  signal dmaparamdataout : std_logic_vector(15 downto 0);
  signal dmaparamdataincdec : std_logic_vector(15 downto 0);

-- instruction monitor output
  signal dmaok : std_logic; 

-- address decoder outputs
  signal a_intr_mask : std_logic;
  signal a_intr_src : std_logic;
  signal a_memcpy_ack : std_logic;
  signal a_sta013_ack : std_logic;
  signal a_idexfer_ack : std_logic;
  signal a_memcpy_go : std_logic;
  signal a_sta013_go : std_logic;
  signal a_idexfer_go : std_logic;
  signal a_dram : std_logic;
  signal drampage_mem : std_logic;
  signal sh_mem : std_logic;
  signal a_reset_ide : std_logic;
  signal a_buf_ide : std_logic;
  signal a_ide : std_logic;
  signal reg_access : std_logic;

-- internal memory control
  signal drampage_rd : std_logic;
  signal dmaparam_rd : std_logic;
  signal drampage_we_lsb : std_logic;
  signal drampage_we_msb : std_logic;
  signal dmaparam_we_lsb : std_logic;
  signal dmaparam_we_msb : std_logic;

-- state machine outputs
  signal done_rd     : std_logic;
  signal done_wr     : std_logic;
  signal done_idexfer : std_logic;
  signal done_memcpy : std_logic;
  signal done_sta013 : std_logic;
  signal all_rows    : std_logic;
  signal refreshdone : std_logic;
  signal use_row     : std_logic;
  signal dma_addr    : std_logic;
  signal paramaddress : std_logic_vector(3 downto 0);
  signal dramcas    : std_logic;
  signal dram_ras    : std_logic;
  signal dramwe     : std_logic;
  signal d16_in_lsb : std_logic;
  signal d16_in_msb : std_logic;
  signal d16_out_lsb : std_logic;
  signal d16_out_msb : std_logic;
  signal d16_rd_lsb : std_logic;
  signal d16_rd_msb : std_logic;
  signal d16_wr_lsb : std_logic;
  signal d16_wr_msb : std_logic;
  signal d16_out_en : std_logic;
  signal ide_addr    : std_logic;
  signal ide_addr_zero : std_logic;
  signal ide_rd      : std_logic;
  signal ide_wr      : std_logic;
  signal register_rd      : std_logic;
  signal register_wr      : std_logic;
  signal sta013dataload     : std_logic;
  signal idedramaddrenable        : std_logic;
  signal rd_en       : std_logic;
  signal incrementparam         : std_logic;
  signal decrementparam         : std_logic;
  signal dmaparam_wr_en   : std_logic;

-- registers output
  signal requestrd : std_logic;
  signal requestwr : std_logic;
  signal requeststa013 : std_logic;
  signal requestmemcpy : std_logic;
  signal requestidexfer : std_logic;
  signal sta013idle : std_logic;

-- ide/dram 
  signal d16_in : std_logic_vector(15 downto 0);
  signal d16_out : std_logic_vector(15 downto 0);
  signal d16_out_prebuf : std_logic_vector(15 downto 0);
  signal d16_tristate : std_logic;
  signal idereset : std_logic;
  signal idedramaddr : std_logic_vector(10 downto 0);
  signal refreshcount : integer range 0 to 224;
  signal requestrefresh : std_logic;
  signal dramweout : std_logic;
  signal ideen_p : std_logic;
  signal ideen_n : std_logic;

-- sta013
  signal sta013request : std_logic;
  
-- dma
  signal countiszero : std_logic;
  
-- components
  component addr_dec
    port (adr: in std_logic_vector(15 downto 0);
          adr_en, reset, clk: in std_logic;
          a_intr_mask, a_intr_src, a_memcpy_ack, a_sta013_ack, a_idexfer_ack,
          a_memcpy_go, a_sta013_go, a_idexfer_go, a_dram, pg_mem, sh_mem,
          a_reset_ide, a_buf_ide, a_ide, reg_access: out std_logic);
  end component;
  component statemachine
    port(requestrd : in std_logic;
         requestwr : in std_logic;
         dma_ok  	: in std_logic;
         a0 : in std_logic;
         requestrefresh : in std_logic;	
         requeststa013 : in std_logic;	
         requestmemcpy : in std_logic;	
         requestidexfer : in std_logic;	
         sta013idle : in std_logic;	
         countiszero : in std_logic;
         a_dram : in std_logic;
         a_ide : in std_logic;
         a_buf_ide : in std_logic;
         clk         : in std_logic;
         reset         : in std_logic;
         done_rd     : out std_logic;
         done_wr     : out std_logic;
         done_idexfer : out std_logic;
         done_memcpy : out std_logic;
         done_sta013 : out std_logic;
         all_rows    : out std_logic;
         done_refresh : out std_logic;
         use_row     : out std_logic;
         dma_addr    : out std_logic;
         paramaddress : out std_logic_vector(3 downto 0);
         dram_cas    : out std_logic;
         dram_ras    : out std_logic;
         dramwe     : out std_logic;
         d16_in_lsb : out std_logic;
         d16_in_msb : out std_logic;
         d16_out_lsb : out std_logic;
         d16_out_msb : out std_logic;
         d16_rd_lsb : out std_logic;
         d16_rd_msb : out std_logic;
         d16_wr_lsb : out std_logic;
         d16_wr_msb : out std_logic;
         d16_out_en : out std_logic;
         ide_addr    : out std_logic;
         ide_addr_zero : out std_logic;
         ide_rd      : out std_logic;
         ide_wr      : out std_logic;
         register_rd      : out std_logic;
         register_wr      : out std_logic;
         sta013dataload     : out std_logic;
         idedramaddrenable        : out std_logic;
         rd_en       : out std_logic;
         incrementparam         : out std_logic;
         decrementparam         : out std_logic;
         shm_wr_en   : out std_logic);
  end component;

  component addresslatch
    port(din : in std_logic_vector(7 downto 0);
         ale : in std_logic;
         clk : in std_logic;
         address_raw : in std_logic_vector(15 downto 8);
         reset : in std_logic;
         adr_en : out std_logic;
         adr : out std_logic_vector(15 downto 0));
  end component;

  component registers
    port(cpuwr : in std_logic;
         cpurd: in std_logic;
         a_idexfer_go : in std_logic;
         a_sta013_go : in std_logic;
         a_memcpy_go : in std_logic;
         a_reset_ide : in std_logic;
         
         done_wr : in std_logic;
         done_rd : in std_logic;
         done_idexfer : in std_logic;
         done_sta013 : in std_logic;
         done_memcpy : in std_logic;

         register_rd : in std_logic;
         register_wr : in std_logic;

         clk : in std_logic;
         reset : in std_logic;

         a_idexfer_ack : in std_logic;
         a_sta013_ack : in std_logic;
         a_memcpy_ack : in std_logic;
         a_intr_mask : in std_logic;
         a_intr_src : in std_logic;
         dmaok : in std_logic;
         requestrefresh : in std_logic;
         
         cpudatain : in std_logic_vector(7 downto 0);
         cpudataout : out std_logic_vector(7 downto 0);

         idereset : out std_logic;
         
         requestrd : out std_logic;
         requestwr : out std_logic;
         requestidexfer : out std_logic;
         requeststa013 : out std_logic;
         requestmemcpy : out std_logic;
         interrupt : out std_logic);
  end component;

  component muxram
    port(din : in std_logic_vector(15 downto 0);
         address_a, address_b  : in std_logic_vector(3 downto 0);
         we_lsb, we_msb, wclk : in std_logic;
         reg_access, dma_addr : in std_logic;
         a0 : in std_logic;
         dout_en : in std_logic;
         dout : out std_logic_vector(7 downto 0);
         q : out std_logic_vector(15 downto 0));
  end component;

  component idedramaddress
    port(clk : in std_logic;
         reset : in std_logic;
         memoryaddress : in std_logic_vector(15 downto 0);
         drampagedata : in std_logic_vector(15 downto 0);
         use_row : in std_logic;
         ide_addr : in std_logic;
         ide_addr_zero : in std_logic;
         all_rows : in std_logic;
         dram_ras : in std_logic;
         idedramaddrenable : in std_logic;
         dramrowaddress : out std_logic_vector(3 downto 0);
         idedramaddr : out std_logic_vector(10 downto 0));
  end component;

  component instructionmonitor
    port (clk : in std_logic;
          reset : in std_logic;
          psen : in std_logic;
          rddone : in std_logic;
          wrdone : in std_logic;
          cpudatain : in std_logic_vector(7 downto 0);
          dmaok : out std_logic);
  end component;

  component sta013driver
    port (clk : in std_logic;
          reset : in std_logic;
          d16_in : in std_logic_vector(15 downto 0);
          sta013dataload : in std_logic;
          sta013request : in std_logic;
          sta013data : out std_logic;
          sta013clock : out std_logic;
          sta013idle : out std_logic);
  end component;
  
-- replace with STARTUP
  component ROC
    port(O : out std_ulogic);
  end component;

  component OFDTX
    port( O : out STD_ULOGIC;
          D : in STD_ULOGIC;
          C : in STD_ULOGIC;
          CE : in STD_ULOGIC;
          T : in STD_ULOGIC);
  end component;

  component IBUF
    port(O : out STD_ULOGIC;
         I : in STD_ULOGIC);
  end component;

  component OBUFT
    port(O : out STD_ULOGIC;
         I : in STD_ULOGIC;
         T : in STD_ULOGIC);
  end component;

  component IFDX
    port(Q : out STD_ULOGIC;
         D : in STD_ULOGIC;
         C : in STD_ULOGIC;
         CE : in STD_ULOGIC);
  end component;

  component IFD
    port(Q : out STD_ULOGIC;
         D : in STD_ULOGIC;
         C : in STD_ULOGIC);
  end component;

  component TDO   
    port(O : in STD_ULOGIC);
  end component;

  component OBUF
    port(O : out STD_ULOGIC;
         I : in STD_ULOGIC);
  end component;

begin

  machine : statemachine port map (requestrd => requestrd,    
                                   requestwr => requestwr,	
                                   dma_ok => dmaok, 
                                   a0 => memoryaddress(0),
                                   requestrefresh => requestrefresh, 
                                   requeststa013 => requeststa013, 
                                   requestmemcpy => requestmemcpy, 
                                   requestidexfer => requestidexfer,
                                   sta013idle => sta013idle,
                                   countiszero   => countiszero,
                                   a_dram   => a_dram,
                                   a_ide    => a_ide,
                                   a_buf_ide => a_buf_ide,
                                   clk         => clk,
                                   reset       => reset,
                                   done_rd     => done_rd,
                                   done_wr     => done_wr,
                                   done_idexfer => done_idexfer,
                                   done_memcpy => done_memcpy,
                                   done_sta013 => done_sta013,
                                   all_rows    => all_rows,
                                   done_refresh => refreshdone,
                                   use_row     => use_row,
                                   dma_addr    => dma_addr,
                                   paramaddress => paramaddress,
                                   dram_cas => dramcas,
                                   dram_ras    => dram_ras,
                                   dramwe     => dramwe,
                                   d16_in_lsb => d16_in_lsb,
                                   d16_in_msb => d16_in_msb,
                                   d16_out_lsb => d16_out_lsb,
                                   d16_out_msb => d16_out_msb,
                                   d16_rd_lsb => d16_rd_lsb,
                                   d16_rd_msb => d16_rd_msb,
                                   d16_wr_lsb => d16_wr_lsb,
                                   d16_wr_msb => d16_wr_msb,
                                   d16_out_en => d16_out_en,
                                   ide_addr    => ide_addr,
                                   ide_addr_zero => ide_addr_zero,
                                   ide_rd      => ide_rd,
                                   ide_wr      => ide_wr,
                                   register_rd      => register_rd,
                                   register_wr      => register_wr,
                                   sta013dataload     => sta013dataload,
                                   idedramaddrenable        => idedramaddrenable,
                                   rd_en       => rd_en,
                                   incrementparam         => incrementparam,
                                   decrementparam         => decrementparam,
                                   shm_wr_en   => dmaparam_wr_en);

  decoder : addr_dec port map (adr => cpuaddress,
                               adr_en => adr_en,
                               reset => reset,
                               clk => clk,
                               a_intr_mask => a_intr_mask,
                               a_intr_src => a_intr_src,
                               a_memcpy_ack => a_memcpy_ack,
                               a_sta013_ack => a_sta013_ack,
                               a_idexfer_ack => a_idexfer_ack,
                               a_memcpy_go => a_memcpy_go,
                               a_sta013_go => a_sta013_go,
                               a_idexfer_go => a_idexfer_go,
                               a_dram => a_dram,
                               pg_mem => drampage_mem,
                               sh_mem => sh_mem,
                               a_reset_ide => a_reset_ide,
                               a_buf_ide => a_buf_ide,
                               a_ide => a_ide,
                               reg_access => reg_access);

  latch : addresslatch port map (din => cpudatain,
                                 ale => ale,   
                                 clk => clk,
                                 address_raw => cpuaddressmsb,
                                 reset => reset,
                                 adr_en => adr_en,
                                 adr => cpuaddress);

  reg : registers port map (cpuwr => cpuwr,
                            cpurd => cpurd,
                            a_idexfer_go => a_idexfer_go,
                            a_sta013_go => a_sta013_go,
                            a_memcpy_go => a_memcpy_go,
                            done_wr => done_wr,
                            done_rd => done_rd,
                            done_idexfer => done_idexfer,
                            done_sta013 => done_sta013,
                            done_memcpy => done_memcpy,
                            clk => clk,
                            requestrd => requestrd,
                            requestwr => requestwr,
                            requestidexfer => requestidexfer,
                            requeststa013 => requeststa013,
                            requestmemcpy => requestmemcpy,
                            reset => reset,
                            cpudatain => cpudatain,
                            cpudataout => cpudataout,
                            register_rd => register_rd,
                            register_wr => register_wr,
                            a_reset_ide => a_reset_ide,
                            idereset => idereset,
                            a_idexfer_ack => a_idexfer_ack,
                            a_sta013_ack => a_sta013_ack,
                            a_memcpy_ack => a_memcpy_ack,
                            a_intr_mask => a_intr_mask,
                            a_intr_src => a_intr_src,
                            dmaok => dmaok,
                            requestrefresh => requestrefresh,
                            interrupt => interrupt);

  pagemem : muxram port map (din(7 downto 0) => cpudatain,
                             din(15 downto 8) => cpudatain,
                             address_a => memoryaddress(4 downto 1),
                             address_b => memoryaddress(15 downto 12),
                             we_lsb => drampage_we_lsb,
                             we_msb => drampage_we_msb,
                             wclk => clk,
                             reg_access => reg_access,
                             dma_addr => dma_addr,
                             a0 => memoryaddress(0),
                             dout_en => drampage_rd,
                             dout => cpudataout,
                             q => drampagedata);

  dmaparam : muxram port map (din(15 downto 0) => dmaparamdatain,
                              address_a => cpuaddress(4 downto 1),
                              address_b => paramaddress,
                              we_lsb => dmaparam_we_lsb,
                              we_msb => dmaparam_we_msb,
                              wclk => clk,
                              reg_access => reg_access,
                              dma_addr => dma_addr,
                              a0 => memoryaddress(0),
                              dout_en => dmaparam_rd,
                              dout => cpudataout,
                              q => dmaparamdataout);

  addressmux : idedramaddress port map (clk => clk,
                                        reset => reset,
                                        memoryaddress => memoryaddress,
                                        drampagedata => drampagedata,
                                        use_row => use_row,
                                        ide_addr => ide_addr,
                                        ide_addr_zero => ide_addr_zero,
                                        all_rows => all_rows,
                                        dram_ras => dram_ras,
                                        idedramaddrenable => idedramaddrenable,
                                        dramrowaddress => dramrowaddress,
                                        idedramaddr => idedramaddr);
  
  instructionmonitor1 :  instructionmonitor port map (clk => clk,
                                                      reset => reset,
                                                      psen => psen,
                                                      rddone => done_rd,
                                                      wrdone => done_wr,
                                                      cpudatain => cpudatain,
                                                      dmaok => dmaok);

  sta013driver1 : sta013driver port map (clk => clk,
                                         reset => reset,
                                         d16_in => d16_in,
                                         sta013dataload => sta013dataload,
                                         sta013request => sta013request,
                                         sta013data => sta013data,
                                         sta013clock => sta013clock,
                                         sta013idle => sta013idle);

  --GLOBAL SET/RESET on configuration
  gsr1 : ROC port map (O => reset);

  dramweoutTDO : TDO port map (O => dramwetdo);
  dramweoutOBUF : OBUF port map (O => dramwetdo, I => dramweout);

  -- cpu bus output drivers
  cpubusout : for n in 7 downto 0 generate
    cpubusout_n : OFDTX port map (O => cpudata(n), D => cpudataout(n), C => clk, T => cpurd, CE => rd_en);
  end generate cpubusout;
  
  -- cpu bus input drivers
  cpubusin : for n in 7 downto 0 generate
    cpubusin_n : IBUF port map (I=> cpudata(n), O => cpudatain(n));
  end generate cpubusin;

  -- IDE/DRAM bus output drivers
  idedrambusout : for n in 15 downto 0 generate
    idedrambusout_n  : OBUFT port map (O => idedramdata(n), I => d16_out_prebuf(n), T => d16_tristate);
  end generate idedrambusout;

  -- IDE/DRAM bus input drivers
  idedrambusin0 : for n in 7 downto 0 generate
    idedrambusin0_n  : IFDX port map (Q => d16_in(n),  D => idedramdata(n), CE => d16_in_lsb, C => clk);
  end generate idedrambusin0;
  idedrambusin1 : for n in 15 downto 8 generate
    idedrambusin1_n  : IFDX port map (Q => d16_in(n),  D => idedramdata(n), CE => d16_in_msb, C => clk);
  end generate idedrambusin1;

  -- IDE/DRAM address
  idedramaddressout : for n in 10 downto 0 generate
    idedramaddressout_n : OBUFT port map (O => idedramaddrout(n), I => idedramaddr(n), T => ideen_n);
  end generate idedramaddressout;

  -- sta013 request
  request : IFD port map(D => sta013requestin, Q => sta013request, C => clk);
  
  -- dram and ide control
  ideen1 : process(clk, ideen, reset)
  begin
    if reset = '1' then
      ideen_p <= '0';
      ideen_n <= '1';
    elsif clk'event and clk = '1' then
      ideen_p <= ideen;
      ideen_n <= not ideen;
    end if;
  end process ideen1;

  iderd <= not ide_rd when ideen_p = '1' else 'Z';
  idewr <= not ide_wr when ideen_p = '1' else 'Z';
  ideresetpin <= idereset when ideen_p = '1' else 'Z';
  dramcasout <= not dramcas;
  dramweout <= not dramwe;

  -- dram refresh timer
  refresh1 : process(clk, reset, refreshcount)
  begin
    if reset = '1' or refreshcount = 224 then
      refreshcount <= 0;
    else
      if clk'event and clk = '1' then
        refreshcount <= refreshcount + 1;
      end if;
    end if;
  end process refresh1;
  refresh2 : process(clk, reset, refreshdone, refreshcount)
  begin
    if reset = '1' or refreshdone = '1' then
      requestrefresh <= '0';
    else
      if clk'event and clk = '1' then
        if refreshcount = 0 then
          requestrefresh <= '1';
        end if;
      end if;
    end if;
  end process refresh2;
  
  -- ide/dram data multiplexing
  d16_out(7 downto 0) <= cpudatain when d16_wr_lsb = '1' else d16_in(7 downto 0);
  d16_out(15 downto 8) <= cpudatain when d16_wr_msb = '1' else d16_in(15 downto 8);
  cpudataout <= d16_in(7 downto 0) when d16_rd_lsb = '1' else (others => 'Z');
  cpudataout <= d16_in(15 downto 8) when d16_rd_msb = '1' else (others => 'Z');
  d16_tristate <= '1' when d16_out_en = '0' or ideen_n = '1' else '0';
  p1 : process(clk, reset, d16_out, d16_out_lsb, d16_out_msb)
  begin
    if reset = '1' then
      d16_out_prebuf <= (others => '0');
      -- rising edge
    else
      if clk'event and clk = '1' then
        if d16_out_lsb = '1' then
          d16_out_prebuf(7 downto 0) <= d16_out(7 downto 0);
        end if;
        if d16_out_msb = '1' then
          d16_out_prebuf(15 downto 8) <= d16_out(15 downto 8);
        end if;
      end if;
    end if;
  end process p1;
  
  -- address source mux, dma parameter memory or cpu address
  memoryaddress <= dmaparamdataout when dma_addr = '1' else cpuaddress;

  -- internal memory decoding
  drampage_rd <= register_rd and drampage_mem;
  dmaparam_rd <= register_rd and sh_mem;
  drampage_we_lsb <= register_wr and drampage_mem and not memoryaddress(0);
  drampage_we_msb <= register_wr and drampage_mem and memoryaddress(0);
  dmaparam_we_lsb <= (register_wr and sh_mem and not cpuaddress(0)) or dmaparam_wr_en;
  dmaparam_we_msb <= (register_wr and sh_mem and cpuaddress(0)) or dmaparam_wr_en;

  -- dma param data
  dmaparamdatain(15 downto 8) <= cpudatain(7 downto 0) when dma_addr = '0' and reg_access = '1' else dmaparamdataincdec(15 downto 8);
  dmaparamdatain(7 downto 0) <= cpudatain(7 downto 0) when dma_addr = '0' and reg_access = '1' else dmaparamdataincdec(7 downto 0);
  dmaparamdataincdec <= dmaparamdataout + 1 when incrementparam = '1' else dmaparamdataout - 1 when decrementparam = '1' else dmaparamdataout;
  countiszero <= '1' when dmaparamdataincdec = "0000000000000000" else '0';
  
end behavior; 
