--
-- MP3 Player, VHDL version (http://www.wiresncode.com/projects/pjrcmp3/fpga)
-- Copyright (c) 2000-1, PJRC.COM, LLC (paul_at_pjrc.com)
-- Copyright (c) 2002, Tom Parker (tparker_at_netspace.net.au)
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_arith.all;
use IEEE.numeric_std.all;

entity registers is 
  port(cpuwr : in std_logic;
       cpurd : in std_logic;
       a_idexfer_go : in std_logic;
       a_sta013_go : in std_logic;
       a_memcpy_go : in std_logic;
       a_reset_ide : in std_logic;
       
       done_wr : in std_logic;
       done_rd : in std_logic;
       done_idexfer : in std_logic;
       done_sta013 : in std_logic;
       done_memcpy : in std_logic;

       register_rd : in std_logic;
       register_wr : in std_logic;

       clk : in std_logic;
       reset : in std_logic;

       a_idexfer_ack : in std_logic;
       a_sta013_ack : in std_logic;
       a_memcpy_ack : in std_logic;
       a_intr_mask : in std_logic;
       a_intr_src : in std_logic;
       dmaok : in std_logic;
       requestrefresh : in std_logic;
       
       cpudatain : in std_logic_vector(7 downto 0);
       cpudataout : out std_logic_vector(7 downto 0);

       idereset : out std_logic;
       
       requestrd : out std_logic;
       requestwr : out std_logic;
       requestidexfer : out std_logic;
       requeststa013 : out std_logic;
       requestmemcpy : out std_logic;
       interrupt : out std_logic);
end registers;

architecture behavior of registers is
  signal rd_phase1 : std_logic;
  signal rd_phase2 : std_logic;
  signal wr_phase1 : std_logic;
  signal wr_phase2 : std_logic;

  signal ideresetdata : std_logic;

  signal idexferinterruptenable : std_logic;
  signal sta013interruptenable : std_logic;
  signal memcpyinterruptenable : std_logic;
  signal dmaokmonitorenable : std_logic;
  signal requestrefreshmonitorenable : std_logic;

  signal idexferinterruptactive : std_logic;
  signal sta013interruptactive : std_logic;
  signal memcpyinterruptactive : std_logic;
  
begin
  -- read falling edge detector
  p1: process(clk, cpurd, done_rd, rd_phase1, rd_phase2, reset)
  begin
    -- rising edge
    if reset = '1' then
      rd_phase1 <= '0';
      rd_phase2 <= '0';
    elsif clk'event and clk = '1' then
      rd_phase1 <= cpurd;
      rd_phase2 <= rd_phase1;
    end if;

    -- falling edge
    if done_rd = '1' or reset = '1' then
      requestrd <= '0';
    elsif clk'event and clk = '0' then
      if rd_phase2 = '1' and rd_phase1 = '0' then
        requestrd <= '1';
      else
        requestrd <= '0';
      end if;
    end if;
  end process p1;

  -- write falling edge detector
  p2: process(clk, cpuwr, done_wr, wr_phase1, wr_phase2, reset)
  begin
    -- rising edge
    if reset = '1' then
      wr_phase1 <= '0';
      wr_phase2 <= '0';
    elsif clk'event and clk = '1' then
      wr_phase1 <= cpuwr;
      wr_phase2 <= wr_phase1;
    end if;

    -- falling edge
    if done_wr = '1' or reset = '1' then
      requestwr <= '0';
    elsif clk'event and clk = '0' then
      if wr_phase2 = '1' and wr_phase1 = '0' then
        requestwr <= '1';
      else
        requestwr <= '0';
      end if;
    end if;
  end process p2;

  -- ide transfer requests
  ide: process(clk, register_wr, reset, a_idexfer_go, done_idexfer)
  begin
    if done_idexfer = '1' or reset = '1' then
      requestidexfer <= '0';
    elsif clk'event and clk = '1' then
      if a_idexfer_go = '1' and register_wr = '1' then
        requestidexfer <= '1';
      end if;
    end if;
  end process ide;
  
  -- sta013 transfer requests
  sta013: process(clk, register_wr, reset, a_sta013_go, done_sta013)
  begin
    if done_sta013 = '1' or reset = '1' then
      requeststa013 <= '0';
    elsif clk'event and clk = '1' then
      if a_sta013_go = '1' and register_wr = '1' then
        requeststa013 <= '1';
      end if;
    end if;
  end process sta013;
  
  -- memcpy transfer requests
  memcpy: process(clk, register_wr, reset, a_memcpy_go, done_memcpy)
  begin
    if done_memcpy = '1' or reset = '1' then
      requestmemcpy <= '0';
    elsif clk'event and clk = '1' then
      if a_memcpy_go = '1' and register_wr = '1' then
        requestmemcpy <= '1';
      end if;
    end if;
  end process memcpy;

  -- ide reset control
  idereset <= ideresetdata;
  cpudataout <= "0000000" & ideresetdata when register_rd = '1' and a_reset_ide = '1' else (others => 'Z');
  ideresetcontrol : process(clk, reset, a_reset_ide, register_wr, cpudatain(0))
  begin
    if reset = '1' then
      ideresetdata <= '0';
    else
      if clk'event and clk = '1' then
        if (a_reset_ide = '1' and register_wr = '1') then
          ideresetdata <= cpudatain(0);
        end if;
      end if;
    end if;
  end process ideresetcontrol;

  -- interrupt masks
  interruptmask : process(clk, reset, register_wr, a_intr_mask, cpudatain)
  begin
    if reset = '1' then
      idexferinterruptenable <= '0';
      sta013interruptenable <= '0';
      memcpyinterruptenable <= '0';
      dmaokmonitorenable <= '0';
      requestrefreshmonitorenable <= '0';
    elsif clk'event and clk = '1' then
      if register_wr = '1' and a_intr_mask = '1' then
        idexferinterruptenable <= cpudatain(0);
        sta013interruptenable <= cpudatain(1);
        memcpyinterruptenable <= cpudatain(2);
        dmaokmonitorenable <= cpudatain(3);
        requestrefreshmonitorenable <= cpudatain(4);
      end if;
    end if;
  end process interruptmask;

  -- interrupt sources (MSB identifies the FPGA)
  cpudataout <= "10000" & memcpyinterruptactive & sta013interruptactive & idexferinterruptactive when register_rd = '1' and a_intr_src = '1' else (others => 'Z');
  ideinterruptsource : process(clk, reset, register_wr, done_idexfer, a_idexfer_ack)
  begin
    if reset = '1' or (register_wr = '1' and a_idexfer_ack = '1') then
      idexferinterruptactive <= '0';
    elsif clk'event and clk = '0' then
      if done_idexfer = '1' then
        idexferinterruptactive <= '1';
      end if;
    end if;
  end process ideinterruptsource;
  sta013interruptsource : process(clk, reset, register_wr, done_sta013, a_sta013_ack)
  begin
    if reset = '1' or (register_wr = '1' and a_sta013_ack = '1') then
      sta013interruptactive <= '0';
    elsif clk'event and clk = '0' then
      if done_sta013 = '1' then
        sta013interruptactive <= '1';
      end if;
    end if;
  end process sta013interruptsource;
  memcpyinterruptsource : process(clk, reset, register_wr, done_memcpy, a_memcpy_ack)
  begin
    if reset = '1' or (register_wr = '1' and a_memcpy_ack = '1') then
      memcpyinterruptactive <= '0';
    elsif clk'event and clk = '0' then
      if done_memcpy = '1' then
        memcpyinterruptactive <= '1';
      end if;
    end if;
  end process memcpyinterruptsource;

  -- interrupt generation (active low)
  interrupt <= '0'      when (idexferinterruptactive = '1' and idexferinterruptenable = '1')
               else '0' when (sta013interruptactive = '1' and sta013interruptenable = '1')
               else '0' when (memcpyinterruptactive = '1' and memcpyinterruptenable = '1')
               else '1';

end behavior; 

