--
-- MP3 Player, VHDL version (http://www.wiresncode.com/projects/pjrcmp3/fpga)
-- Copyright (c) 2000-1, PJRC.COM, LLC (paul_at_pjrc.com)
-- Copyright (c) 2002, Tom Parker (tparker_at_netspace.net.au)
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_arith.all;

entity statemachine is
  port(

    -- INPUTS

    requestrd      : in std_logic;
    requestwr      : in std_logic;

    dma_ok      : in std_logic;

    a0 : in std_logic;

    requestrefresh : in std_logic; 
    requeststa013  : in std_logic; 
    requestmemcpy  : in std_logic; 
    requestidexfer : in std_logic; 

    sta013idle : in std_logic;
    
    countiszero   : in std_logic;

    a_dram   : in std_logic;
    a_ide    : in std_logic;
    a_buf_ide : in std_logic;
    clk         : in std_logic;
    reset       : in std_logic;
    
    -- OUTPUTS

    done_rd     : out std_logic;
    done_wr     : out std_logic;
    done_idexfer : out std_logic;
    done_memcpy : out std_logic;
    done_sta013 : out std_logic;

    all_rows    : out std_logic;
    done_refresh : out std_logic;
    use_row     : out std_logic;
    dma_addr    : out std_logic;
    paramaddress : out std_logic_vector(3 downto 0);
    dram_cas    : out std_logic;
    dram_ras    : out std_logic;
    dramwe     : out std_logic;

    ide_addr    : out std_logic;
    ide_addr_zero : out std_logic;
    ide_rd      : out std_logic;
    ide_wr      : out std_logic;

    d16_in_lsb : out std_logic;
    d16_in_msb : out std_logic;
    d16_out_lsb : out std_logic;
    d16_out_msb : out std_logic;
    d16_rd_lsb : out std_logic;
    d16_rd_msb : out std_logic;
    d16_wr_lsb : out std_logic;
    d16_wr_msb : out std_logic;
    d16_out_en : out std_logic;
    
    register_rd      : out std_logic;
    register_wr      : out std_logic;

    sta013dataload     : out std_logic;

    idedramaddrenable        : out std_logic;
    rd_en       : out std_logic;

    incrementparam         : out std_logic;
    decrementparam         : out std_logic;

    shm_wr_en   : out std_logic
    );
end statemachine;

architecture behaviour of statemachine is

-- misc
  signal state_change : std_logic;
  signal done_read    : std_logic;
  signal done_write   : std_logic;
  signal dma_addr_on  : std_logic;
  signal ca0_on       : std_logic;
  signal ca2_on       : std_logic;

-- derived commands
  signal do_read      : std_logic;
  signal do_write     : std_logic;
  signal do_wait      : std_logic;
  signal do_refresh   : std_logic;
  signal do_sta013    : std_logic;
  signal do_memcpy    : std_logic;
  signal do_idexfer   : std_logic;
  signal do_nothing   : std_logic;
  signal iszero      : std_logic;

-- single state
  signal s_rd_reg     : std_logic;
  signal s_wr_reg     : std_logic;
  signal s_memcpy     : std_logic;
  signal s_idexfer_end : std_logic;
  signal s_memcpy_end : std_logic;
  signal s_sta013_start : std_logic;
  signal s_sta013_end : std_logic;

-- multiple state machines
  -- DRAM Read
  -- 1: RAS fetch
  -- 2: RAS output, Column fetch
  -- 3: Column output
  -- 4: Latch 16 bit data from DRAM
  -- 5: Latch 8 bit data from 16 bit latch to CPU bus
  signal s_rd_dram    : std_logic_vector(5 downto 1);
  -- DRAM Write
  signal s_wr_dram    : std_logic_vector(7 downto 1);
  signal s_rd_ide     : std_logic_vector(6 downto 1);
  signal s_wr_ide     : std_logic_vector(6 downto 1);
  -- IDE DMA Read
  -- 1: ide address and CS setup (ide_addr_zero)
  -- 2: 
  signal s_idexfer    : std_logic_vector(9 downto 1);
  signal s_sta013     : std_logic_vector(7 downto 1);
  signal s_refresh    : std_logic_vector(9 downto 1);

-- ide/dram data bus control
  signal d16_out_one : std_logic;
  signal d16_in_both : std_logic;
  signal d16_out_both : std_logic;
  signal d16_rd      : std_logic;
  signal d16_wr      : std_logic;
  signal requeststa013ok : std_logic;
  

begin
  -- commands
  do_read <= requestrd and not dma_ok;
  do_write <= requestwr and not dma_ok and not requestrd;
  do_wait <= not requestwr and not dma_ok and not requestrd;
  do_refresh <= requestrefresh and dma_ok;
  do_sta013 <= requeststa013ok and dma_ok and not requestrefresh;
  do_memcpy <= requestmemcpy and dma_ok and not requestrefresh and not requeststa013ok;
  do_idexfer <= requestidexfer and dma_ok and not requestrefresh and not requeststa013ok and not do_memcpy;
  do_nothing <= dma_ok and not requestrefresh and not requeststa013ok and not do_memcpy and not do_idexfer;

  -- sta013 start
  s_sta013_start <= do_sta013 and state_change;
  requeststa013ok <= requeststa013 and sta013idle;
  
  -- misc outputs
  done_rd <= done_read;
  done_wr <= done_write;
  done_memcpy <= s_memcpy_end;
  incrementparam <= s_idexfer(4) or s_idexfer(5) or s_idexfer(7) or s_idexfer(8)
	         or s_sta013(2) or s_sta013(3) or s_sta013(4) or s_sta013(5);
  decrementparam <= s_idexfer(2) or s_idexfer(3) or s_sta013(6) or s_sta013(7);
  idedramaddrenable <= s_rd_dram(1) or s_rd_dram(2) or s_wr_dram(1) or s_wr_dram(2)
          or s_rd_ide(1) or s_wr_ide(1) or s_idexfer(1) or s_idexfer(6)
          or s_idexfer(7) or s_sta013(1) or s_sta013(2);
  -- dataout bus to cpu bus clock enable
  rd_en <= s_rd_reg or s_rd_dram(5) or s_rd_ide(6);
  shm_wr_en <= s_idexfer(3) or s_idexfer(5) or s_idexfer(8) or s_sta013(3)
               or s_sta013(5) or s_sta013(7);
  dma_addr <= dma_addr_on;
  paramaddress(0) <= ca0_on;
  paramaddress(2) <= ca2_on;
  
  -- sta013 outputs
  sta013dataload <= s_sta013(5);
  done_sta013 <= s_sta013_end;

  -- dram outputs
  all_rows <= s_refresh(3);
  use_row <= s_refresh(3) or s_rd_dram(1) or s_wr_dram(1) or s_idexfer(6) or s_sta013(1);
  paramaddress(1) <= s_idexfer(2) or s_idexfer(3) or s_idexfer(9);
  paramaddress(3) <= '0';
  dramwe <= s_wr_dram(6) or s_wr_dram(7) or s_idexfer(8) or s_idexfer(9);
  done_refresh <= s_refresh(5);
  dram_cas <= s_refresh(3) or s_refresh(4) or s_rd_dram(3) or s_rd_dram(4)
              or s_wr_dram(3) or s_wr_dram(4) or s_wr_dram(7) or s_idexfer(9)
              or s_sta013(3) or s_sta013(4);
  dram_ras <= s_refresh(3) or s_refresh(4) or s_rd_dram(2) or s_rd_dram(3)
              or s_rd_dram(4) or s_wr_dram(2) or s_wr_dram(3) or s_wr_dram(4)
              or s_wr_dram(5) or s_wr_dram(6) or s_wr_dram(7) or s_idexfer(7)
              or s_idexfer(8) or s_idexfer(9) or s_sta013(2) or s_sta013(3) 
              or s_sta013(4);

  -- ide/dram data bus control
  d16_rd <=     s_rd_dram(5) or s_rd_ide(6);
  d16_wr <=     s_wr_dram(5) or s_wr_ide(2);
  d16_out_one <= s_wr_ide(2);
  d16_in_both <= s_rd_dram(4) or s_wr_dram(4) or s_rd_ide(5) or s_idexfer(5) or s_sta013(4);
  d16_out_both <= s_wr_dram(5) or s_idexfer(6);
  d16_out_en <= s_wr_dram(6) or s_wr_dram(7) or s_wr_ide(3) or s_wr_ide(4) or s_wr_ide(5)
		or s_wr_ide(6) or s_idexfer(8) or s_idexfer(9);

  d16_in_lsb <= d16_in_both;
  d16_in_msb <= d16_in_both;
  d16_out_lsb <= d16_out_both or ((d16_out_one or (a_buf_ide and s_wr_reg)) and not a0);
  d16_out_msb <= d16_out_both or ((d16_out_one or (a_buf_ide and s_wr_reg)) and a0);
  d16_rd_lsb <= (d16_rd or (a_buf_ide and s_rd_reg)) and not a0;
  d16_rd_msb <= (d16_rd or (a_buf_ide and s_rd_reg)) and a0;
  d16_wr_lsb <= (d16_wr or (a_buf_ide and s_wr_reg)) and not a0;
  d16_wr_msb <= (d16_wr or (a_buf_ide and s_wr_reg)) and a0;
  
  -- ide outputs
  done_idexfer <= s_idexfer_end;
  ide_addr <= s_rd_ide(1) or s_wr_ide(1);
  ide_addr_zero <= s_idexfer(1);
  -- changed to meet setup time
  --ide_wr <= s_wr_ide(3) or s_wr_ide(4) or s_wr_ide(5);
  ide_wr <= s_wr_ide(4) or s_wr_ide(5);
  ide_rd <= s_rd_ide(3) or s_rd_ide(4) or s_rd_ide(5) 
            or s_idexfer(3) or s_idexfer(4) or s_idexfer(5); 

  -- register outputs
  register_rd <= s_rd_reg;
  register_wr <= s_wr_reg;

  -- state changes
  p1: process(do_read,
              do_write,
              do_wait,
              do_idexfer,
              do_nothing,
              do_memcpy,
              do_refresh,
              
              a_dram,
              a_ide,
              clk,
              countiszero,
              reset,
              state_change,
              s_rd_dram,
              s_wr_dram,

              s_refresh,
              s_rd_ide,
              s_wr_ide,
              s_idexfer,
              s_sta013,
              s_sta013_start,

              s_rd_reg,
              s_wr_reg,

              iszero,
              s_memcpy,

              dma_addr_on,
              ca0_on,

              ca2_on,
              done_read,
              done_write,
              s_idexfer_end,

              s_memcpy_end,
              s_sta013_end)
  begin
    -- rising edge
    if reset = '1' then
      iszero <= '0';  
    else if (clk'event and clk = '1') then
           iszero <= countiszero;
         end if;
    end if;

    -- falling edge
    if reset = '1' then
      s_rd_reg <= '0';
      s_wr_reg <= '0';
      s_rd_dram <= (others => '0');
      s_wr_dram <= (others => '0');
      s_refresh <= (others => '0');
      s_rd_ide <= (others => '0');
      s_wr_ide <= (others => '0');
      s_idexfer <= (others => '0');
      s_memcpy <= '0';
      s_sta013 <= (others => '0');
      done_read <= '0';
      done_write <= '0';
      s_idexfer_end <= '0';
      s_memcpy_end <= '0';
      s_sta013_end <= '0';
      dma_addr_on <= '0';
      ca0_on <= '0';
      ca2_on <= '0';
      state_change <= '1';
    else if (clk'event and clk = '0') then
           -- Registers
           s_rd_reg <= do_read and not a_dram and not a_ide and state_change;
           s_wr_reg <= do_write and not a_dram and not a_ide and state_change;
           -- DRAM
           s_rd_dram <= s_rd_dram(4 downto 1) & (do_read and a_dram and state_change);
           s_wr_dram <= s_wr_dram(6 downto 1) & (do_write and a_dram and state_change);
           s_refresh <= s_refresh(8 downto 1) & (do_refresh and state_change);
           -- IDE
           s_rd_ide <= s_rd_ide(5 downto 1) & (do_read and a_ide and state_change);
           s_wr_ide <= s_wr_ide(5 downto 1) & (do_write and a_ide and state_change);
           s_idexfer <= s_idexfer(8 downto 1) & (do_idexfer and state_change);
           -- MEMCPY
           s_memcpy <= do_memcpy and state_change;
           -- STA013
           s_sta013 <= s_sta013(6 downto 1) & s_sta013_start;

           -- DONE
           done_read <= s_rd_reg or s_rd_dram(5) or s_rd_ide(6);
           done_write <= s_wr_reg or s_wr_ide(6) or s_wr_dram(7);
           s_idexfer_end <= s_idexfer(9) and iszero;
           s_memcpy_end <= s_memcpy;  -- TODO
           s_sta013_end <= s_sta013(7) and iszero;

           -- outputs
           dma_addr_on <= not s_idexfer(9) and not s_sta013(7) and (s_sta013_start or s_idexfer(1) or dma_addr_on);
           ca0_on <= not s_idexfer(8) and not s_sta013(7) and (s_sta013(5) or s_idexfer(3) or ca0_on);
           ca2_on <= not s_sta013(7) and (s_sta013_start or ca2_on);


           -- state change
           state_change <= (do_nothing and state_change)
                           or (do_wait and state_change)
                           or done_read
                           or done_write
                           or (not iszero and s_idexfer(9))
                           or s_idexfer_end
                           or s_memcpy_end
                           or (not iszero and s_sta013(7))
                           or s_sta013_end
                           or s_refresh(9);
         end if;
  end if;
end process P1;
end behaviour;


